/*
 * SDLKobo v0.1
----------------------------------------------------------------------
	gfxengine.C - Graphics Engine
----------------------------------------------------------------------
 *  David Olofson, 2001
 * This code is released under the terms of the GNU GPL.
 */

#define	DBG(x)	x

#include <math.h>
#include "SDL.h"
#include "SDL_image.h"
#include "sofont.h"

#include "gfxengine.h"
#include "window.h"
#include "vidsync.h"


gfxengine_t *gfxengine;


gfxengine_t::gfxengine_t()
{
	gfxengine = this;	/* Uurgh! Kludge. */

	screen = NULL;
	window = NULL;
	wx = wy = 0;
	gfx = NULL;
	csengine = NULL;
	buffermode = DOUBLE;
	syncmode = FLIP;
	_fullscreen = 0;
	use_interpolation = 1;
	_depth = 0;
	_title = "GfxEngine v0.1";
	_icontitle = "GfxEngine";
	_cursor = 1;
	_width = 320;
	_height = 240;

	start_tick = 0;
	ticks_per_frame = 1000.0/60.0;

	is_running = 0;
	is_showing = 0;
	is_open = 0;

	memset(fonts, 0, sizeof(fonts));

	xscroll = yscroll = 0;
	for(int i = 0; i < CS_LAYERS ; ++i)
		xratio[i] = yratio[i] = 0.0;
}


gfxengine_t::~gfxengine_t()
{
	stop();
	hide();
	close();
	for(int i = 0; i < GFX_BANKS; ++i)
		delete fonts[i];
}


void gfxengine_t::output(window_t *outwin)
{
	window = outwin;
}


/*----------------------------------------------------------
	Initialization
----------------------------------------------------------*/

void gfxengine_t::size(int w, int h)
{
	int was_showing = is_showing;
	hide();

	_width = w;
	_height = h;
	if(csengine)
		cs_engine_set_size(csengine, w, h);

	if(was_showing)
		show();
}

void gfxengine_t::mode(int bits, int fullscreen)
{
	int was_showing = is_showing;
	hide();

	int olddepth = _depth;
	_depth = bits;
	if(_depth != olddepth)
		reload();
	_fullscreen = fullscreen;

	if(was_showing)
		show();
}

void gfxengine_t::buffering(buffermodes_t bm, syncmodes_t sm)
{
	int was_showing = is_showing;
	hide();

	buffermode = bm;
	syncmode = sm;

	if(was_showing)
		show();
}

void gfxengine_t::period(float frameduration)
{
	if(frameduration > 0)
	{
		DBG(printf("Setting period to %f ms.\n", frameduration);)
		ticks_per_frame = frameduration;
	}
	else
		DBG(printf("Time line reset.\n");)	
	start_tick = SDL_GetTicks();
	if(csengine)
		cs_engine_advance(csengine, 0);
}

void gfxengine_t::wrap(int x, int y)
{
	wx = x;
	wy = y;
	if(csengine)
		cs_engine_set_wrap(csengine, x, y);
}


/*----------------------------------------------------------
	Data management
----------------------------------------------------------*/

void gfxengine_t::colorkey(int r, int g, int b)
{
	r = (r>>8)&0xff;
	g = (g>>8)&0xff;
	b = (b>>8)&0xff;
	s_colorkey = (r<<16) | (g<<8) | b;
}


int gfxengine_t::loadimage(int bank, const char *name)
{
	if(!csengine)
	{
		fprintf(stderr, "loadimage: Engine must be open!\n");
		return -10;
	}
	s_blitmode = S_BLITMODE_NORMAL;
	DBG(printf("Loading image %s (bank %d)... ", name, bank);)
	if(s_load_image(gfx, bank, name))
	{
		fprintf(stderr, "Failed to load %s!\n", name);
		return -1;
	}

	s_bank_t *b = s_get_bank(gfx, bank);
	if(!b)
	{ 
		fprintf(stderr, "gfxengine: Internal error 1!\n");
		return -15;
	}
	cs_engine_set_image_size(csengine, bank, b->w, b->h);

	DBG(printf("Ok.\n");)
}


int gfxengine_t::loadsprites(int bank, int w, int h, const char *name)
{
	if(!csengine)
	{
		fprintf(stderr, "loadsprite: Engine must be open!\n");
		return -10;
	}
	s_blitmode = S_BLITMODE_COLORKEY;
	DBG(printf("Loading sprites %s (bank %d; %dx%d)... ",
			name, bank, w, h);)
	if(s_load_bank(gfx, bank, w, h, name))
	{
		fprintf(stderr, "Failed to load %s!\n", name);
		return -2;
	}

	cs_engine_set_image_size(csengine, bank, w, h);

	DBG(printf("Ok. (%d frames)\n", s_get_bank(gfx, bank)->max+1);)
}


int gfxengine_t::loadfont(int bank, const char *name)
{
	if(!csengine)
	{
		fprintf(stderr, "loadfont: Engine must be open!\n");
		return -10;
	}
	s_blitmode = S_BLITMODE_NORMAL;
	DBG(printf("Loading font %s (bank %d)... ", name, bank);)
	if(s_load_image(gfx, bank, name))
	{
		fprintf(stderr, "Failed to load %s!\n", name);
		return -2;
	}
	if(bank >= GFX_BANKS)
	{
		fprintf(stderr, "Too high bank #!\n");
		return -3;
	}

	if(!fonts[bank])
		fonts[bank] = new SoFont;
	if(!fonts[bank])
	{
		fprintf(stderr, "Failed to instantiate SoFont!\n");
		return -4;
	}

	if(fonts[bank]->load(s_get_sprite(gfx, bank, 0)->surface))
	{
		DBG(printf("Ok.\n");)
		return 0;
	}
	else
	{
		fprintf(stderr, "SoFont::load() failed!\n");
		return -5;
	}
}


void gfxengine_t::reload()
{
	DBG(printf("Reloading all banks. (Not implemented!)\n");)
//	s_reload_all_banks(gfx);
}


void gfxengine_t::unload(int bank)
{
	DBG(printf("Unloading bank %d.\n", bank);)
	s_delete_bank(gfx, bank);
}


/*----------------------------------------------------------
	Engine open/close
----------------------------------------------------------*/

void gfxengine_t::on_frame(cs_engine_t *e)
{
	gfxengine->__frame();
	gfxengine->frame();
}


int gfxengine_t::open(int objects)
{
	if(is_open)
		return 0;

	DBG(printf("Opening engine...\n");)
	csengine = cs_engine_create(_width, _height, objects);
	if(!csengine)
	{
		fprintf(stderr, "Failed to set up control system engine!\n");
		return -1;
	}

	csengine->on_frame = on_frame;
	cs_engine_set_wrap(csengine, wx, wy);

	gfx = s_new_container(GFX_BANKS);
	if(!gfx)
	{
		fprintf(stderr, "Failed to set up graphics container!\n");
		cs_engine_delete(csengine);
		return -2;
	}

	is_open = 1;
	return show();
}


void gfxengine_t::close()
{
	if(!is_open)
		return;

	DBG(printf("Closing engine...\n");)
	stop();
	hide();
	cs_engine_delete(csengine);
	csengine = NULL;
	s_delete_container(gfx);
	gfx = NULL;
	SDL_Quit();
	is_open = 0;
}


/*----------------------------------------------------------
	Settings
----------------------------------------------------------*/

void gfxengine_t::title(const char *win, const char *icon)
{
	_title = win;
	_icontitle = icon;
	if(screen)
		SDL_WM_SetCaption(_title, _icontitle);
}


/*----------------------------------------------------------
	Display show/hide
----------------------------------------------------------*/

int gfxengine_t::show()
{
	int flags = SDL_HWSURFACE;

	if(is_showing)
		return 0;

	DBG(printf("Opening screen...\n");)
	if(SDL_Init(SDL_INIT_VIDEO) == -1)
	{
		fprintf(stderr, "Failed to initialize SDL!\n");
		return -2;
	}

	switch(buffermode)
	{
	  case SINGLE:
		break;
	  case DOUBLE:
		flags |= SDL_DOUBLEBUF;
		break;
	  case HALF:
		flags &= ~SDL_HWSURFACE;
		break;
	}

	if(_fullscreen)
		flags |= SDL_FULLSCREEN;

	screen = SDL_SetVideoMode(_width, _height, _depth, flags);
	if(!screen)
	{
		fprintf(stderr, "Failed to open display!\n");
		return -3;
	}

	SDL_WM_SetCaption(_title, _icontitle);
	SDL_ShowCursor(_cursor);
	cs_engine_set_size(csengine, _width, _height);
	csengine->filter = use_interpolation;

	is_showing = 1;
	return 0;
}


void gfxengine_t::hide(void)
{
	if(!is_showing)
		return;

	DBG(printf("Closing screen...\n");)
	stop();
	screen = NULL;
	SDL_Quit();

	is_showing = 0;
}


/*----------------------------------------------------------
	Engine start/stop
----------------------------------------------------------*/

void gfxengine_t::start_engine()
{
	if(syncmode == RETRACE)
		vidsync_init(0.3);
	start_tick = SDL_GetTicks();
}


void gfxengine_t::stop_engine()
{
	if(syncmode == RETRACE)
		vidsync_close();
}


/*----------------------------------------------------------
	Control
----------------------------------------------------------*/

/*
 * Run engine until stop() is called.
 *
 * The virtual member frame() will be called once for
 * each control system frame. pre_render() and
 * post_render() will be called before/after the engine
 * renders each video frame.
 */
void gfxengine_t::run()
{
	open();
	show();
	start_engine();
	is_running = 1;
	while(is_running)
	{
		int tick = SDL_GetTicks() - start_tick;
		float toframe = (float)tick / ticks_per_frame;
		cs_engine_advance(csengine, toframe);
		pre_render();
		window->select();
		cs_engine_render(csengine);
		post_render();
		flip();
	}
	stop_engine();
}


void gfxengine_t::stop()
{
	if(!is_running)
		return;

	DBG(printf("Stopping engine...\n");)
	is_running = 0;
}


void gfxengine_t::cursor(int csr)
{
	_cursor = csr;
	if(screen)
		SDL_ShowCursor(csr);
}


void gfxengine_t::interpolation(int inter)
{
	use_interpolation = inter;
	if(!csengine)
		return;

	csengine->filter = use_interpolation;
}


void gfxengine_t::scroll_ratio(int layer, float xr, float yr)
{
	if(layer < 0)
		return;
	if(layer >= CS_LAYERS)
		return;

	xratio[layer] = xr;
	yratio[layer] = yr;
}


void gfxengine_t::scroll(int xs, int ys)
{
	xscroll = xs;
	yscroll = ys;

	/* Apply current scroll pos to layers */
	for(int i = 0; i < CS_LAYERS ; ++i)
	{
		csengine->offsets[i].v.x = (int)floor(xscroll * xratio[i]);
		csengine->offsets[i].v.y = (int)floor(yscroll * yratio[i]);
	}
}


void gfxengine_t::force_scroll()
{
	if(csengine)
		for(int i = 0; i < CS_LAYERS ; ++i)
			cs_point_force(&csengine->offsets[i]);
}


int gfxengine_t::xoffs(int layer)
{
	if(layer < 0)
		return 0;
	if(layer >= CS_LAYERS)
		return 0;
	return csengine->offsets[layer].gx;
}


int gfxengine_t::yoffs(int layer)
{
	if(layer < 0)
		return 0;
	if(layer >= CS_LAYERS)
		return 0;
	return csengine->offsets[layer].gy;
}


/*----------------------------------------------------------
	Internal stuff
----------------------------------------------------------*/

/* Default frame handler */
void gfxengine_t::frame()
{
	SDL_Event ev;
	while(SDL_PollEvent(&ev))
	{
		if(ev.type == SDL_KEYDOWN)
			if(ev.key.keysym.sym == SDLK_ESCAPE)
				stop();
	}
}


/* Internal frame handler */
void gfxengine_t::__frame()
{
}


void gfxengine_t::pre_render()
{
}


void gfxengine_t::post_render()
{
}


void gfxengine_t::sync_half()
{
	switch(syncmode)
	{
	  case FLIP:
		/* Not applicable. */
		break;
	  case TIMER:
		/* Not yet implemented. */
		break;
	  case RETRACE:
#ifdef linux
		vidsync_wait(0.47);
#endif
		break;
	}
}

void gfxengine_t::sync()
{
	switch(syncmode)
	{
	  case FLIP:
		/* Do nothing. */
		break;
	  case TIMER:
		/* Not yet implemented. */
		break;
	  case RETRACE:
#ifdef linux
		vidsync_wait(0);
#endif
		break;
	}
}

void gfxengine_t::flip()
{
	switch(buffermode)
	{
	  case SINGLE:
		sync();
		SDL_UpdateRect(screen, 0,0, 0,0);
		break;
	  case DOUBLE:
		sync();
		SDL_Flip(screen);
		break;
	  case HALF:
		sync_half();
		SDL_UpdateRect(screen, 0,0, 320,120);
		sync();
		SDL_UpdateRect(screen, 0,120, 320,120);
		break;
	}
}


/*
 * Generic render() callback for sprites and tiles.
 */
void gfxengine_t::render_sprite(cs_obj_t *o)
{
	SDL_Rect dest_rect;
	SDL_Surface *img;
	img = gfxengine->get_sprite(o->anim.bank, o->anim.frame);
	if(!img)
		return;

	dest_rect.x = CS2PIXEL(o->point.gx) + gfxengine->window->x();
	dest_rect.y = CS2PIXEL(o->point.gy) + gfxengine->window->y();
	SDL_BlitSurface(img, NULL, gfxengine->screen, &dest_rect);
}


cs_obj_t *gfxengine_t::get_obj(int layer)
{
	cs_obj_t *o = cs_engine_get_obj(csengine);
	if(o)
	{
		o->render = render_sprite;
		cs_obj_layer(o, layer);
		cs_obj_activate(o);
	}
	return o;
}


void gfxengine_t::free_obj(cs_obj_t *obj)
{
	cs_obj_free(obj);
}


int gfxengine_t::objects_in_use()
{
	if(!csengine)
		return 0;

	return csengine->pool_total - csengine->pool_free;
}


SDL_Surface *gfxengine_t::get_sprite(int bank, int frame)
{
	s_sprite_t *s = s_get_sprite(gfx, bank, frame);
	if(s)
		return s->surface;
	else
		return NULL;
}


SoFont *gfxengine_t::get_font(unsigned int f)
{
	if(f < GFX_BANKS)
		return fonts[f];
	else
		return NULL;
}
