/*
 * Simple proof-of-concept retrace sync hack by
 * David Olofson, 2001.
 */

#include <sys/io.h>

#include "vidsync.h"

#include <stdio.h>
#define ERROUT(x...)				\
{						\
	fprintf(stderr, "vidsync ERROR: ");	\
	fprintf(stderr, ## x);			\
}

#if 0
#define DBGOUT(x...)				\
{						\
	fprintf(stderr, "vidsync: ");		\
	fprintf(stderr, ## x);			\
}
#else
#define DBGOUT(x...)
#endif


static int late_margin;
static int no_syncs;

/*
 * "Milliframe" timing functions
 * (Actually 128 units/frame in this version.)
 */

long long mft_clocks_per_frame = 128;
long long mft_last_reset = 0;

__inline__ unsigned long long int rdtsc(void)
{
	unsigned long long int x;
	__asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
	return x;
}

/*
 * Reset the virtual raster counter
 */
void mftime_reset(void)
{
	mft_last_reset = rdtsc();
}

/*
 * Tweak the virtual raster counter in the direction
 * of the current time by a specified amount.
 * "amount" is in the usual 7 bit fraction format; ie
 * 128 == full on.
 */
void mftime_sync(int amount)
{
	long long diff = rdtsc() - mft_last_reset;
	diff %= mft_clocks_per_frame;
	if(diff > mft_clocks_per_frame >> 1)
		diff -= mft_clocks_per_frame;
	diff *= (long long)amount;
	mft_last_reset += diff / 128LL;
	DBGOUT("sync %f\n", (float)diff/mft_clocks_per_frame/128LL);
}

/*
 * Push the virtual raster counter forward
 */
void mftime_adjust(int diff)
{
	mft_last_reset += (long long)diff * (mft_clocks_per_frame >> 7);
}

/*
 * Return the current time in milliframes since last reset
 */
int mftime_get(void)
{
	return (rdtsc() - mft_last_reset) / (mft_clocks_per_frame >> 7);
}

/*
 * The code used for VGA compatible cards;
 * ie, if no special function is provided.
 * This should work on most cards.
 */

static void hw_retrace_sync(void)
{
	while (!(inb(IS1_RC) & 8));
	while (inb(IS1_RC) & 8);
}

static int hw_retrace_sync_timeout(int timeout)
{
	while (!(inb(IS1_RC) & 8))
		if(mftime_get() > timeout)
			return -1;
	while (inb(IS1_RC) & 8)
		if(mftime_get() > timeout)
			return -2;
	return 0;
}

/*
 * Calibrate the timer to the current refresh rate
 */
#define	UNREALISTIC_COUNT	100000000000
int mftime_init(void)
{
	long long time1, time2;
	long long last;
	int i;
	mft_clocks_per_frame = UNREALISTIC_COUNT;
	for(i=0; i<5; ++i)
	{
		hw_retrace_sync();
		time1 = rdtsc();
		hw_retrace_sync();
		time2 = rdtsc();
		last = time2 - time1;
		if(last < mft_clocks_per_frame)
			mft_clocks_per_frame = last;
	}
	mftime_reset();
	no_syncs = 0;
	if(mft_clocks_per_frame != UNREALISTIC_COUNT)
		DBGOUT("mftime_init() done; "
				"clocks_per_frame = %lld\n", mft_clocks_per_frame);
	else
	{
		ERROUT("mftime_init() failed!\n");
		return -42;
	}
	return 0;
}


int vidsync_init(float tolerance)
{
	int res = ioperm(IS1_RC, 1, 1);
	if(0 != res)
	{
		ERROUT("ioperm() failed!\n");
		return res;
	}
	res = mftime_init();
	late_margin = 128.0 * tolerance;
	DBGOUT("Initialized.\n");
	return res;
}

void vidsync_close(void)
{
	ioperm(IS1_RC, 1, 0);
	DBGOUT("Closed.\n");
}

void vidsync_wait_retrace()
{
	int fraction, frames;

	fraction = mftime_get();

	/*
	 * Calculate frames + fractional frame.
	 */
	frames = fraction >> 7;
	fraction &= 0x7f;
	if(fraction > late_margin)
		fraction -= 128;

	/*
	 * Forget any retraces that passed while we were away.
	 */
	mftime_adjust(frames << 7);

	no_syncs += frames;

	/*
	 * Handle many lost frames as an error condition, and recalibrate.
	 */
	if( (frames > VIDSYNC_MAX_LOST_FRAMES)
				||
			(no_syncs > VIDSYNC_FRAMES_NO_SYNC) )
	{
		ERROUT("Missed %d frames and "
			"no retrace sync for %d frames; "
			" recalibrating...\n", frames - 1, no_syncs);
		mftime_init();
		return;
	}
	else if(frames > 1)
		DBGOUT("Missed %d frames.\n", frames - 1);

	/*
	 * Don't bother polling if we're late.
	 */
	if(fraction > -VIDSYNC_POLL_MARGIN)
	{
		DBGOUT("Late vidsync_wait() call; not polling."
				" fraction was %d.\n", fraction);
		return;
	}

	/*
	 * Ok, we're here early, so we do a real
	 * retrace sync, and then reset the timer.
	 */
	if(hw_retrace_sync_timeout(128 + late_margin) == 0)
	{
		/*
		 * Only reset the timer on a _real_ retrace
		 * sync!
		 */
		mftime_sync(8);
		no_syncs = 0;
	}
	else
		DBGOUT("vidsync_wait() polling timed out. "
				" fraction was %d.\n", fraction);
}

void vidsync_wait(float position)
{
	if(position <= 0.0)
		vidsync_wait_retrace();
	else
		while(mftime_get() < position * 128.0)
			;
}
